import 'dart:async';
import 'dart:ui' as ui;
import 'package:image/image.dart' as img;
import 'package:flutter/widgets.dart';
import 'package:xcommonwidgets/xcommonwidgets.dart';
import 'package:xfxwidgets/xMediaItem.dart';
import 'package:xtoolspackage/xtoolspackage.dart';

import './ImagePainter_ColorItem.dart';
import './ImagePainter_Controller.dart';
import './ImagePainter_RangeSlider.dart';
import './ImagePainter_SelectionItems.dart';
import './ImagePainter_TextDelegate.dart';
import './ImagePainter_TextDialog.dart';
import './ImagePainter_Class.dart';
import 'package:flutter/material.dart' hide Image;
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';

///[ImagePainter] widget.
@immutable
class ImagePainter extends StatefulWidget {
  ImagePainter._({
    Key? key,
    this.height,
    this.width,
    this.placeHolder,
    this.isScalable,
    this.brushIcon,
    this.clearAllIcon,
    this.colorIcon,
    this.loadMedia,
    this.undoIcon,
    this.controlsAtTop = true,
    this.signatureBackgroundColor = Colors.white,
    this.colors,
    this.initialPaintMode,
    this.initialStrokeWidth,
    this.initialColor,
    this.onColorChanged,
    this.onStrokeWidthChanged,
    this.onPaintModeChanged,
    this.textDelegate,
    this.xMedia,
    this.backImage,
  }) : super(key: key);

  ///Constructor for loading image from network url.
  factory ImagePainter.network(
    String url, {
    required Key key,
    double? height,
    double? width,
    Widget? placeholderWidget,
    bool? scalable,
    List<Color>? colors,
    Widget? brushIcon,
    Widget? undoIcon,
    Widget? clearAllIcon,
    Widget? colorIcon,
    PaintMode? initialPaintMode,
    double? initialStrokeWidth,
    Color? initialColor,
    ValueChanged<PaintMode>? onPaintModeChanged,
    ValueChanged<Color>? onColorChanged,
    ValueChanged<double>? onStrokeWidthChanged,
    TextDelegate? textDelegate,
    bool? controlsAtTop,
  }) {
    return ImagePainter._(
      key: key,
      height: height,
      width: width,
      placeHolder: placeholderWidget,
      isScalable: scalable,
      colors: colors,
      brushIcon: brushIcon,
      undoIcon: undoIcon,
      colorIcon: colorIcon,
      clearAllIcon: clearAllIcon,
      initialPaintMode: initialPaintMode,
      initialColor: initialColor,
      initialStrokeWidth: initialStrokeWidth,
      onPaintModeChanged: onPaintModeChanged,
      onColorChanged: onColorChanged,
      onStrokeWidthChanged: onStrokeWidthChanged,
      textDelegate: textDelegate,
      controlsAtTop: controlsAtTop ?? true,
    );
  }

  ///Constructor for loading image from assetPath.

  ///Constructor for signature painting.
  factory ImagePainter.signature({
    required Key key,
    Color? signatureBgColor,
    double? height,
    double? width,
    List<Color>? colors,
    Widget? brushIcon,
    Widget? undoIcon,
    Widget? clearAllIcon,
    Widget? colorIcon,
    bool isNota = false,
    bool isNotaWithImage = false,
    ValueChanged<PaintMode>? onPaintModeChanged,
    ValueChanged<Color>? onColorChanged,
    ValueChanged<double>? onStrokeWidthChanged,
    TextDelegate? textDelegate,
    bool? controlsAtTop,
    Future Function(dynamic)? loadMedia,
    XMediaItem? xMedia,
  }) {
    return ImagePainter._(
      key: key,
      height: height,
      width: width,
      xMedia: xMedia,
      isScalable: false,
      colors: colors,
      loadMedia: loadMedia,
      signatureBackgroundColor: signatureBgColor ?? Colors.white,
      brushIcon: brushIcon,
      undoIcon: undoIcon,
      colorIcon: colorIcon,
      clearAllIcon: clearAllIcon,
      onPaintModeChanged: onPaintModeChanged,
      onColorChanged: onColorChanged,
      onStrokeWidthChanged: onStrokeWidthChanged,
      textDelegate: textDelegate,
      controlsAtTop: controlsAtTop ?? true,
    );
  }

  // ///Only accessible through [ImagePainter.network] constructor.
  // final String? networkUrl;

  // ///Only accessible through [ImagePainter.memory] constructor.
  // final Uint8List? byteArray;

  // ///Only accessible through [ImagePainter.file] constructor.
  // final File? file;

  // ///Only accessible through [ImagePainter.asset] constructor.
  // final String? assetPath;

  XMediaItem? xMedia;
  Image? backImage;

  ///Height of the Widget. Image is subjected to fit within the given height.
  double? height;

  ///Width of the widget. Image is subjected to fit within the given width.
  double? width;
  final Future Function(dynamic)? loadMedia;

  ///Widget to be shown during the conversion of provided image to [ui.Image].
  final Widget? placeHolder;

  ///Defines whether the widget should be scaled or not. Defaults to [false].
  final bool? isScalable;

  ///Signature mode background color
  final Color signatureBackgroundColor;

  ///List of colors for color selection
  ///If not provided, default colors are used.
  List<Color>? colors = null;

  ///Icon Widget of strokeWidth.
  final Widget? brushIcon;

  ///Widget of Color Icon in control bar.
  final Widget? colorIcon;

  ///Widget for Undo last action on control bar.
  final Widget? undoIcon;

  ///Widget for clearing all actions on control bar.
  final Widget? clearAllIcon;

  ///Define where the controls is located.
  ///`true` represents top.
  final bool controlsAtTop;

  ///Initial PaintMode.
  final PaintMode? initialPaintMode;

  //the initial stroke width
  final double? initialStrokeWidth;

  //the initial color
  final Color? initialColor;

  final ValueChanged<Color>? onColorChanged;

  final ValueChanged<double>? onStrokeWidthChanged;

  final ValueChanged<PaintMode>? onPaintModeChanged;

  //the text delegate
  final TextDelegate? textDelegate;

  @override
  ImagePainterState createState() => ImagePainterState();
}

///
class ImagePainterState extends State<ImagePainter> {
  final _repaintKey = GlobalKey();
  ui.Image? _image;
  late Controller _controller;
  late final ValueNotifier<bool> _isLoaded;
  late final TextEditingController _textController;
  late final TransformationController _transformationController_Payload;
  late final TransformationController _transformationController_Overlay;
  //final _imagePainterState_Key = GlobalKey<ImagePainterState>();

  int _strokeMultiplier = 1;
  late TextDelegate textDelegate;

  bool modifiedActive = false;
  bool isBusy = true;

  @override
  void initState() {
    super.initState();
    if (widget.colors == null) {
      widget.colors = List<Color>.empty(growable: true);
      widget.colors!.addAll([Colors.yellow, Colors.black, Colors.red, Colors.white, Colors.purple, Colors.blue, Colors.green]);
    }
    _isLoaded = ValueNotifier<bool>(false);
    _controller = Controller();

    loadMediaAndImage();
  }

  loadMediaAndImage() async {
    //carichiamo i payloads se uno è vuoto
    if (widget.xMedia!.payload == null) {
      await widget.loadMedia!(widget.xMedia);
    } else {}

    if (widget.xMedia!.payload != null) {
      widget.backImage = Image.memory(
        widget.xMedia!.payload!,
        // height: MediaQuery.of(context).size.height / 2,
        // width: MediaQuery.of(context).size.width / 2,
      );
    }
    // var ii = widget.backImage.image as MemoryImage;

    if (widget.xMedia!.overlaySurface == null) {
      if (widget.xMedia!.payload != null) {
        _image = await createTransparentImage_fromPayLoad(widget.xMedia!.payload!);
      } else {
        _image = await createTransparentImage(widget.width!.toInt(), widget.height!.toInt());
      }
    } else {
      _image = await _convertTo_UiImage(widget.xMedia!.overlaySurface!);
    }
    if (_image == null) {
      throw ("Image couldn't be resolved from xMedia!.overlaySurface.");
    } else {
      isBusy = false;
      _setStrokeMultiplier();
    }

    _isLoaded.value = true;

    _controller.naturalSize = imageSize;
    //if (widget.isSignature) {
    _controller.update(
      mode: PaintMode.none,
      color: Colors.black,
    );
    // } else {
    //   _controller.update(
    //     mode: widget.initialPaintMode,
    //     strokeWidth: widget.initialStrokeWidth,
    //     color: widget.initialColor,
    //   );
    // }
    _textController = TextEditingController();
    _transformationController_Payload = TransformationController();
    _transformationController_Overlay = TransformationController();
    textDelegate = widget.textDelegate ?? TextDelegate();
    _isLoaded.value = true;
  }

  Future<ui.Image> imgImageToUiImage(img.Image image) async {
    //https://medium.com/@ys.commerciale/process-and-show-an-image-in-flutter-aebb0054ce94
    ui.ImmutableBuffer buffer = await ui.ImmutableBuffer.fromUint8List(image.getBytes());
    ui.ImageDescriptor id = ui.ImageDescriptor.raw(buffer, height: image.height, width: image.width, pixelFormat: ui.PixelFormat.rgba8888);
    ui.Codec codec = await id.instantiateCodec(targetHeight: image.height, targetWidth: image.width);
    ui.FrameInfo fi = await codec.getNextFrame();
    ui.Image uiImage = fi.image;

    return uiImage;
  }

  ///Completer function to convert asset or file image to [ui.Image] before drawing on custompainter.
  Future<ui.Image> _convertTo_UiImage(Uint8List bytes) async {
    final completer = Completer<ui.Image>();
    ui.decodeImageFromList(bytes, (image) {
      return completer.complete(image);
    });
    return completer.future;
  }

  Future<ui.Image> createTransparentImage_fromPayLoad(Uint8List payloadOri) async {
    var sz = await imageSize_from(payloadOri);
    var i = await createTransparentImage(sz.width.toInt(), sz.height.toInt());
    return i;
  }

  Future<ui.Image> createTransparentImage(int width, int height) {
    final img.Image image = img.Image(
      width: width,
      height: height,
      numChannels: 4,
      format: img.Format.uint8, //uint8
    );
    final completer = Completer<ui.Image>();

    ui.decodeImageFromPixels(image.buffer.asUint8List(), image.width, image.height, ui.PixelFormat.rgba8888, (image) {
      return completer.complete(image);
    });
    return completer.future;
  }

  // Future<ui.Image> createTransparentUiImage(int width, int height) async {
  //   final img.Image image = img.Image(
  //     width: width,
  //     height: height,
  //     numChannels: 4,
  //     format: img.Format.uint8, //uint8
  //   );
  //   final completer = Completer<ui.Image>();
  // }

  ///Dynamically sets stroke multiplier on the basis of widget size.
  ///Implemented to avoid thin stroke on high res images.
  _setStrokeMultiplier() {
    if ((_image!.height + _image!.width) > 1000) {
      _strokeMultiplier = (_image!.height + _image!.width) ~/ 1000;
    }
    _controller.update(strokeMultiplier: _strokeMultiplier);
  }
//      var mi = MemoryImage(bytes);
// ui.Image i = await mi.loadImage(mi.obtainKey(configuration), (buffer, {getTargetSize}) => null);
//     return Image.memory(bytes);
//     // final completer = Completer<Image>();
//     // ui.decodeImageFromList(img, (image) {
//     //   return completer.complete(image);
//     // });
//     // return completer.future;
//   }

  Future<Size> imageSize_from(Uint8List bytes) async {
    // final bytes = Uint8List.fromList([
    //   // PNG image bytes here
    // ]);

    final codec = await ui.instantiateImageCodec(bytes);
    final frame = await codec.getNextFrame();
    final size = Size(frame.image.width.toDouble(), frame.image.height.toDouble());

    return size;
    // final completer = Completer<Image>();
    // ui.decodeImageFromList(img, (image) {
    //   return completer.complete(image);
    // });
    // return completer.future;
  }

  @override
  void dispose() {
    _controller.dispose();
    _isLoaded.dispose();
    _textController.dispose();
    _transformationController_Payload.dispose();
    _transformationController_Overlay.dispose();
    super.dispose();
  }

  bool get isEdited => _controller.paintHistory.isNotEmpty;

  Size get imageSize => Size(_image == null ? widget.width ?? double.maxFinite : _image?.width.toDouble() ?? 0, _image == null ? widget.height ?? double.maxFinite : _image?.height.toDouble() ?? 0);

  ///Completer function to convert network image to [ui.Image] before drawing on custompainter.
  Future<ui.Image> _loadNetworkImage(String path) async {
    final completer = Completer<ImageInfo>();
    final img = NetworkImage(path);
    img.resolve(const ImageConfiguration()).addListener(ImageStreamListener((info, _) => completer.complete(info)));
    final imageInfo = await completer.future;
    _isLoaded.value = true;
    return imageInfo.image;
  }

  @override
  Widget build(BuildContext context) {
    return ValueListenableBuilder<bool>(
      valueListenable: _isLoaded,
      builder: (_, loaded, __) {
        if (loaded) {
          return _paintSignature();
          //return widget.isSignature ? _paintSignature() : _paintImage();
        } else {
          return Container(
            height: widget.height ?? double.maxFinite,
            width: widget.width ?? double.maxFinite,
            child: Center(
              child: widget.placeHolder ?? const CircularProgressIndicator(),
            ),
          );
        }
      },
    );
  }

  Widget _paintSignature() {
    print(_controller.naturalSize);
    return Container(
        height: MediaQuery.of(context).size.height / 1.2,
        child: StatefulBuilder(builder: (context, setState) {
          return XAlertDialog(
              child: Stack(children: [
                Container(
                    color: Colors.white,
                    height: widget.height ?? double.maxFinite,
                    width: widget.width ?? double.maxFinite,
                    child: Column(children: [
                      modifiedActive
                          ? widget.controlsAtTop
                              ? _buildControls()
                              : Container()
                          : Container(),
                      Expanded(
                          child: FittedBox(
                              alignment: FractionalOffset.topLeft,
                              child: RepaintBoundary(
                                  key: _repaintKey,
                                  child: ClipRect(
                                      child: AnimatedBuilder(
                                          animation: _controller,
                                          builder: (context, child) {
                                            return InteractiveViewer(
                                                transformationController: _transformationController_Overlay,
                                                maxScale: 4,
                                                minScale: 0.02,
                                                panEnabled: _controller.mode == PaintMode.none,
                                                scaleEnabled: _controller.mode == PaintMode.none,
                                                onInteractionStart: _scaleStartGesture,
                                                onInteractionUpdate: _scaleUpdateGesture,
                                                onInteractionEnd: _scaleEndGesture,
                                                child: Stack(alignment: Alignment.topLeft, children: [
                                                  widget.xMedia!.payload != null
                                                      ? Container(
                                                          //margin: EdgeInsets.only(top: 20, left: 10),
                                                          //////child: Image.memory(widget.xMedia!.payload!,
                                                          child: widget.backImage,
                                                        )
                                                      : Container(),
                                                  CustomPaint(
                                                      size: Size(_controller.naturalSize.width, _controller.naturalSize.height),
                                                      willChange: true,
                                                      isComplex: true,
                                                      painter: DrawImage(
                                                        image: _image,
                                                        backgroundColor: Colors.transparent,
                                                        isSignature: true,
                                                        controller: _controller,
                                                      )),
                                                ]));
                                          }))))),
                      modifiedActive
                          ? widget.controlsAtTop.not()
                              ? _buildControls()
                              : Container()
                          : Container(),
                      //SizedBox(height: MediaQuery.of(context).padding.bottom)
                    ])),
                XStatefulWidgetBase.LoaderIndicator(isBusy, label: "... elaborazione in corso ...", color_Waiter: Colors.amber.shade500)
              ]),
              content_insetPadding: EdgeInsets.symmetric(horizontal: 4),
              actionsBTNarea_Padding: EdgeInsets.all(0),
              actionsBTNarea_Child: Row(children: [
                modifiedActive.not()
                    ? Expanded(
                        child: XBtnbase(
                        height: 40,
                        width: 120,
                        backGroundColor: Colors.transparent,
                        onPressed: () {
                          setState(() {
                            modifiedActive = true;
                            _controller.setMode(PaintMode.freeStyle);
                          });
                        },
                        elevation: 0,
                        label: "Annota",
                        label_Style: TextStyle(color: Colors.orange[600], fontSize: 18),
                      ))
                    : Container(),
              ]));
        }));
  }

  cmd_Save() async {
    if (_image != null) {
      widget.xMedia!.overlaySurface = await exportImage();
      widget.xMedia!.miniatura = await exportFullThumbnails();
      // var x = createThumbnails_fromImage(widget.xMedia!.payload!);
      // if (x != null) {
      //   widget.xMedia!.miniatura = createThumbnailsWithOverlay_img(widget.xMedia!.overlaySurface!, x);
      //   print("miniatura len ${widget.xMedia!.miniatura!.length}");
      // } else
      //   print("miniatura NOT DONE!");
    }
  }

  Uint8List? createThumbnails_fromImage(Uint8List image_payload, {int width = 80, int height = 100}) {
    var imgOri = img.decodePng(image_payload);
    if (imgOri == null) imgOri = img.decodeJpg(image_payload);
    if (imgOri == null) imgOri = img.decodeBmp(image_payload);
    if (imgOri == null) imgOri = img.decodeImage(image_payload);
    if (imgOri == null) imgOri = img.decodeExr(image_payload);
    if (imgOri == null) imgOri = img.decodeGif(image_payload);
    if (imgOri == null) return null;

    var resizedImg = img.copyResize(imgOri, width: width, height: height);
    final pngBytes = img.encodePng(resizedImg);
    return pngBytes;
  }

  Uint8List createThumbnailsWithOverlay_img(Uint8List overlay_payload, Uint8List image_Thumbnails, {Size? size}) {
    print("createThumbnailsWithOverlay START");
    var s = Stopwatch();
    s.start();
    int w = 0;
    int h = 0;
    final overLayImg = img.decodePng(overlay_payload, frame: 1)!;
    s.stop();
    print("createThumbnailsWithOverlay: img.decodePng(overlay_payload)${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
    s.start();

    if (size != null) // allineo vert/hor all'overlay
    {
      w = size.width.toInt();
      h = size.height.toInt();

      if (overLayImg.width > overLayImg.height) {
        w = size.height.toInt();
        ;
        h = size.width.toInt();
      }
    } else {
      w = overLayImg.width ~/ 10;
      h = overLayImg.height ~/ 10;
    }
    var resizedOverlayImg = img.copyResize(overLayImg, width: w, height: h);
    s.stop();
    print("createThumbnailsWithOverlay:  img.copyResize()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
    s.start();
    var imgThumb = img.decodePng(image_Thumbnails)!;
    s.stop();
    print("createThumbnailsWithOverlay:  img.decodePng(image_Thumbnails)${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
    s.start();

    var newMiniature = img.compositeImage(imgThumb, resizedOverlayImg, blend: img.BlendMode.alpha);
    s.stop();
    print("createThumbnailsWithOverlay:  img.compositeImage()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
    s.start();

    final pngBytes = img.encodePng(newMiniature);
    s.stop();
    print("createThumbnailsWithOverlay:  img.copyResize()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
    // s.stop();
    // print("createThumbnailsWithOverlay:${s.elapsedMicroseconds} for ${overLayImg..lengthInBytes}");
    return pngBytes;
  }

//   Future<Uint8List> createThumbnailsWithOverlay_ui(Uint8List overlay_payload, Uint8List image_Thumbnails, {Size? size}) async {
//     print("createThumbnailsWithOverlay START");
//     var s = Stopwatch();
//     s.start();
//     int w = 0;
//     int h = 0;
//     //final overLayImg = img.decodePng(overlay_payload)!;
//     ui.Codec codec = await ui.instantiateImageCodec(overlay_payload);
//     ui.FrameInfo frameInfo = await codec.getNextFrame();
//     final overLayImg = frameInfo.image;
// var p =ip.ImagePicker();
// p.
//     s.stop();
//     print("createThumbnailsWithOverlay: img.decodePng(overlay_payload)${s.elapsedMilliseconds}ms for ${overLayImg}");
//     s.start();

//     if (size != null) // allineo vert/hor all'overlay
//     {
//       w = size.width.toInt();
//       h = size.height.toInt();

//       if (overLayImg.width > overLayImg.height) {
//         w = size!.height.toInt();
//         ;
//         h = size!.width.toInt();
//       }
//     } else {
//       w = overLayImg.width ~/ 10;
//       h = overLayImg.height ~/ 10;
//     }

//     var resizedOverlayImg = ui.Image.cop(width: w, height: h);
//     s.stop();
//     print("createThumbnailsWithOverlay:  img.copyResize()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
//     s.start();
//     var imgThumb = img.decodePng(image_Thumbnails)!;
//     s.stop();
//     print("createThumbnailsWithOverlay:  img.decodePng(image_Thumbnails)${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
//     s.start();

//     var newMiniature = img.compositeImage(imgThumb, resizedOverlayImg, blend: img.BlendMode.alpha);
//     s.stop();
//     print("createThumbnailsWithOverlay:  img.compositeImage()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
//     s.start();

//     final pngBytes = img.encodePng(newMiniature);
//     s.stop();
//     print("createThumbnailsWithOverlay:  img.copyResize()${s.elapsedMilliseconds}ms for ${overLayImg..lengthInBytes}");
//     // s.stop();
//     // print("createThumbnailsWithOverlay:${s.elapsedMicroseconds} for ${overLayImg..lengthInBytes}");
//     return pngBytes;
//   }

  _scaleStartGesture(ScaleStartDetails onStart) {
    if (_controller.mode == PaintMode.none) return;

    final _zoomAdjustedOffset = _transformationController_Overlay.toScene(onStart.localFocalPoint);
    _controller.setStart(_zoomAdjustedOffset);
    _controller.addOffsets(_zoomAdjustedOffset);
  }

  ///Fires while user is interacting with the screen to record painting.
  void _scaleUpdateGesture(ScaleUpdateDetails onUpdate) {
    if (_controller.mode == PaintMode.none) return;
    if (widget.xMedia!.isModified.not()) {
      setState(() {
        widget.xMedia!.isModified = true;
      });
    }
    final _zoomAdjustedOffset = _transformationController_Overlay.toScene(onUpdate.localFocalPoint);
    _controller.setInProgress(true);
    if (_controller.start == null) {
      _controller.setStart(_zoomAdjustedOffset);
    }
    _controller.setEnd(_zoomAdjustedOffset);
    if (_controller.mode == PaintMode.freeStyle) {
      _controller.addOffsets(_zoomAdjustedOffset);
    }
    if (_controller.onTextUpdateMode) {
      _controller.paintHistory.lastWhere((element) => element.mode == PaintMode.text).offsets = [_zoomAdjustedOffset];
    }
  }

  ///Fires when user stops interacting with the screen.
  void _scaleEndGesture(ScaleEndDetails onEnd) {
    if (_controller.mode == PaintMode.none) return;

    _controller.setInProgress(false);
    if (_controller.start != null && _controller.end != null && (_controller.mode == PaintMode.freeStyle)) {
      _controller.addOffsets(null);
      _addFreeStylePoints();
      _controller.offsets.clear();
    } else if (_controller.start != null && _controller.end != null && _controller.mode != PaintMode.text) {
      _addEndPoints();
    }
    _controller.resetStartAndEnd();
  }

  void _addEndPoints() => _addPaintHistory(
        PaintInfo(
          offsets: <Offset?>[_controller.start, _controller.end],
          mode: _controller.mode,
          color: _controller.color,
          strokeWidth: _controller.scaledStrokeWidth,
          fill: _controller.fill,
        ),
      );

  void _addFreeStylePoints() => _addPaintHistory(
        PaintInfo(
          offsets: <Offset?>[..._controller.offsets],
          mode: PaintMode.freeStyle,
          color: _controller.color,
          strokeWidth: _controller.scaledStrokeWidth,
        ),
      );

  ///Provides [ui.Image] of the recorded canvas to perform action.
  Future<ui.Image> _renderImage() async {
    final recorder = ui.PictureRecorder();
    final canvas = Canvas(recorder);
    final painter = DrawImage(image: _image, controller: _controller);
    //final size = Size(_image!.width!.toDouble(), _image!.height!.toDouble());
    final size = Size(_controller.naturalSize.width, _controller.naturalSize.height);
    painter.paint(canvas, size);
    return recorder.endRecording().toImage(size.width.floor(), size.height.floor());
  }

  // Future<ui.Image> _renderFullThumbnails() async {
  //   final recorder = ui.PictureRecorder();
  //   final canvas = Canvas(recorder);
  //   final painter = DrawImage(image: _image, controller: _controller);
  //   //final size = Size(_image!.width!.toDouble(), _image!.height!.toDouble());
  //   final size = Size(_controller.naturalSize.width, _controller.naturalSize.height);
  //   painter.paint(canvas, size);

  //   return recorder.endRecording().toImage(size.width.floor(), size.height.floor());
  // }

  PopupMenuItem _showOptionsRow() {
    return PopupMenuItem(
      enabled: false,
      child: Center(
        child: SizedBox(
          child: Wrap(
            children: paintModes(textDelegate)
                .map(
                  (item) => SelectionItems(
                    data: item,
                    isSelected: _controller.mode == item.mode,
                    onTap: () {
                      if (widget.onPaintModeChanged != null) {
                        widget.onPaintModeChanged!(item.mode);
                      }
                      _controller.setMode(item.mode);

                      Navigator.of(context).pop();
                      if (item.mode == PaintMode.text) {
                        _openTextDialog();
                      }
                    },
                  ),
                )
                .toList(),
          ),
        ),
      ),
    );
  }

  PopupMenuItem _showRangeSlider() {
    return PopupMenuItem(
      enabled: false,
      child: SizedBox(
        width: double.maxFinite,
        child: AnimatedBuilder(
          animation: _controller,
          builder: (_, __) {
            return RangedSlider(
              value: _controller.strokeWidth,
              onChanged: (value) {
                _controller.setStrokeWidth(value);
                if (widget.onStrokeWidthChanged != null) {
                  widget.onStrokeWidthChanged!(value);
                }
              },
            );
          },
        ),
      ),
    );
  }

  PopupMenuItem _showColorPicker() {
    return PopupMenuItem(
      enabled: false,
      child: Center(
        child: Wrap(
          alignment: WrapAlignment.center,
          spacing: 10,
          runSpacing: 10,
          children: (widget.colors ?? editorColors).map((color) {
            return ColorItem(
              isSelected: color == _controller.color,
              color: color,
              onTap: () {
                _controller.setColor(color);
                if (widget.onColorChanged != null) {
                  widget.onColorChanged!(color);
                }
                Navigator.pop(context);
              },
            );
          }).toList(),
        ),
      ),
    );
  }

  ///Generates [Uint8List] of the [ui.Image] generated by the [renderImage()] method.
  ///Can be converted to image file by writing as bytes.
  Future<Uint8List?> exportImage() async {
    late ui.Image _convertedImage;
    //  questo fa un mix della foto in back + l'overlay
    // final _boundary = _repaintKey.currentContext!.findRenderObject() as RenderRepaintBoundary;
    // _convertedImage = await _boundary.toImage(pixelRatio: 3);

    //a noi interessa SOLO _image (ns overlay puro)
    _convertedImage = await _renderImage();
    final byteData = await _convertedImage.toByteData(format: ui.ImageByteFormat.png);
    return byteData?.buffer.asUint8List();
  }

  Future<Uint8List?> exportFullThumbnails() async {
    late ui.Image _convertedImage;
    //  questo fa un mix della foto in back + l'overlay
    final _boundary = _repaintKey.currentContext!.findRenderObject() as RenderRepaintBoundary;
    _convertedImage = await _boundary.toImage(pixelRatio: 0.05);

    final byteData = await _convertedImage.toByteData(format: ui.ImageByteFormat.png);
    return byteData?.buffer.asUint8List();
  }

  void _addPaintHistory(PaintInfo info) {
    if (info.mode != PaintMode.none) {
      _controller.addPaintInfo(info);
    }
  }

  void _openTextDialog() {
    _controller.setMode(PaintMode.text);
    final fontSize = 6 * _controller.strokeWidth;
    TextDialog.show(
      context,
      _textController,
      fontSize,
      _controller.color,
      textDelegate,
      onFinished: (context) {
        if (_textController.text.isNotEmpty) {
          _addPaintHistory(
            PaintInfo(
              mode: PaintMode.text,
              text: _textController.text,
              offsets: [],
              color: _controller.color,
              strokeWidth: _controller.scaledStrokeWidth,
            ),
          );
          _textController.clear();
        }
        Navigator.of(context).pop();
      },
    );
  }

  Widget _buildControls() {
    return Container(
      padding: const EdgeInsets.all(4),
      color: Colors.grey[200],
      child: Row(
        children: [
          AnimatedBuilder(
            animation: _controller,
            builder: (_, __) {
              final icon = paintModes(textDelegate).firstWhere((item) => item.mode == _controller.mode).icon;
              return PopupMenuButton(
                tooltip: textDelegate.changeMode,
                shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(40)),
                icon: Icon(icon, color: Colors.black),
                itemBuilder: (_) => [_showOptionsRow()],
              );
            },
          ),
          AnimatedBuilder(
            animation: _controller,
            builder: (_, __) {
              return PopupMenuButton(
                padding: const EdgeInsets.symmetric(vertical: 10),
                shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(20)),
                tooltip: textDelegate.changeColor,
                icon: widget.colorIcon ??
                    Container(
                      padding: const EdgeInsets.all(2.0),
                      decoration: BoxDecoration(
                        shape: BoxShape.circle,
                        border: Border.all(color: Colors.black),
                        color: _controller.color,
                      ),
                    ),
                itemBuilder: (_) => [_showColorPicker()],
              );
            },
          ),
          PopupMenuButton(
            tooltip: textDelegate.changeBrushSize,
            shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(20)),
            icon: widget.brushIcon ?? Icon(Icons.brush, color: Colors.black),
            itemBuilder: (_) => [_showRangeSlider()],
          ),
          // AnimatedBuilder(
          //   animation: _controller,
          //   builder: (_, __) {
          //     if (_controller.canFill()) {
          //       return Row(children: [
          //         Checkbox.adaptive(
          //             value: _controller.shouldFill,
          //             checkColor: Colors.black,
          //             onChanged: (val) {
          //               _controller.update(fill: val);
          //             }),
          //         Text('Riempi', style: TextStyle(color: Colors.black))
          //       ]);
          //     } else {
          //       return const SizedBox();
          //     }
          //   },
          // ),
          const Spacer(),

          IconButton(
            tooltip: textDelegate.undo,
            icon: widget.undoIcon ?? Icon(Icons.reply, color: Colors.black),
            onPressed: () => _controller.undo(),
          ),
          IconButton(
            tooltip: textDelegate.clearAllProgress,
            icon: widget.clearAllIcon ?? Icon(Icons.delete_forever_outlined, color: Colors.black),
            onPressed: () => _controller.clear(),
          ),
          IconButton(
            tooltip: "Premi per avere più carta",
            icon: Icon(Icons.image_aspect_ratio, color: Colors.black),
            onPressed: () => _controller.morePaper(),
          ),
          widget.xMedia!.isModified
              ? XBtnbase(
                  height: 40,
                  width: 40,
                  toolTip_Message: "Salva le tue modifiche ed Esci",
                  icon: Icons.done,
                  icon_Color: Colors.green,
                  backGroundColor: Colors.transparent,
                  onPressed: () async {
                    modifiedActive = false;
                    setState(() {
                      isBusy = true;
                    });
                    setState(() {
                      _transformationController_Overlay.value = Matrix4.identity();
                    });
                    await cmd_Save();
                    setState(() {
                      isBusy = false;
                      widget.xMedia!.isModified = false;
                    });
                    modifiedActive = false;
                    Navigator.pop(context, true);
                  },
                  elevation: 0,
                )
              : Container(),
          XBtnbase(
            height: 40,
            width: 40,
            backGroundColor: Colors.transparent,
            onPressed: () {
              modifiedActive = false;
              widget.xMedia!.isModified = false;
              Navigator.pop(context, false);
            },
            elevation: 0,
            toolTip_Message: "Chiudi l'editor",
            icon: Icons.close,
            icon_Color: Colors.red,
          ),
        ],
      ),
    );
  }
}

// // ///paints image on given constrains for drawing if image is not null.
// // Widget _paintImage() {
// //   return Container(
// //     height: widget.height ?? double.maxFinite,
// //     width: widget.width ?? double.maxFinite,
// //     child: Column(
// //       children: [
// //         if (widget.controlsAtTop) _buildControls(),
// //         Expanded(
// //           child: FittedBox(
// //             alignment: FractionalOffset.center,
// //             child: ClipRect(
// //               child: AnimatedBuilder(
// //                 animation: _controller,
// //                 builder: (context, child) {
// //                   return InteractiveViewer(
// //                     transformationController: _transformationController_Overlay,
// //                     maxScale: 2.4,
// //                     minScale: 1,
// //                     panEnabled: _controller.mode == PaintMode.none,
// //                     scaleEnabled: widget.isScalable!,
// //                     onInteractionUpdate: _scaleUpdateGesture,
// //                     onInteractionEnd: _scaleEndGesture,
// //                     child: CustomPaint(
// //                       size: imageSize,
// //                       willChange: true,
// //                       isComplex: true,
// //                       painter: DrawImage(
// //                         image: _image,
// //                         controller: _controller,
// //                       ),
// //                     ),
// //                   );
// //                 },
// //               ),
// //             ),
// //           ),
// //         ),
// //         if (!widget.controlsAtTop) _buildControls(),
// //         SizedBox(height: MediaQuery.of(context).padding.bottom)
// //       ],
// //     ),
// //   );
// // }
