library xfxwidgets;

import 'dart:io';
import 'dart:math';
import 'dart:typed_data';
import 'package:android_intent_plus/android_intent.dart';
import 'package:android_intent_plus/flag.dart';
import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
import 'package:path_provider/path_provider.dart';
import 'package:xcommonwidgets/xcommonwidgets.dart';
import 'package:xfxbase/XBL/XAzureRest.dart';
import 'package:xfxbase/XFDataItem.dart';
import 'package:xfxbase/_XFX/XF_Col.dart';
import 'package:xfxbase/_XFX/XF_View.dart';
import 'package:xtoolspackage/XLibStorage.dart';
import 'package:xtoolspackage/xtoolspackage.dart';
import 'package:flutter_slidable/flutter_slidable.dart';
import 'package:open_file/open_file.dart';

import 'XF_ELENCO/XElenco.dart';
import 'XF_MAIN/XApp.dart';
import 'XF_MAIN/XAppState.dart';
import 'XTabItem.dart';

part 'XF_ELENCO/XElencoItem.dart';
part 'XF_SCHEDA/XfxSchedaTapped.dart';
part 'XF_SCHEDA/XfxScheda.dart';
part 'XfxUpdater.dart';

abstract class XEnums {
  static XEnumMediaType mediaType = XEnumMediaType();
}

class XEnumMediaType<int> extends XEnum {
  XEnumDef image = XEnumDef(value: 0, label: 'Image', backColor: Color.fromARGB(255, 0, 0, 0), textColor: Color.fromARGB(255, 0, 0, 0));
  XEnumDef audio = XEnumDef(value: 1, label: 'Audio', backColor: Color.fromARGB(255, 0, 0, 0), textColor: Color.fromARGB(255, 0, 0, 0));
  XEnumDef video = XEnumDef(value: 2, label: 'Video', backColor: Color.fromARGB(255, 0, 0, 0), textColor: Color.fromARGB(255, 0, 0, 0));
  XEnumDef panoramica = XEnumDef(value: 10, label: 'Panoramica', backColor: Color.fromARGB(255, 0, 0, 0), textColor: Color.fromARGB(255, 0, 0, 0));
  XEnumDef multiImage = XEnumDef(value: 20, label: 'MultiImage', backColor: Color.fromARGB(255, 0, 0, 0), textColor: Color.fromARGB(255, 0, 0, 0));
  @override
  void initDefinitions() {
    definitions[0] = image;
    definitions[1] = audio;
    definitions[2] = video;
    definitions[10] = panoramica;
    definitions[20] = multiImage;
  }
}

abstract class XfxUtils {
  static String GetStringFromValue(dynamic value, {XProp? colForXDataItem = null}) {
    if (value is XFDataItem) {
      var str = value[colForXDataItem!.col_Name].toString();
      return str;
    } else if (value is XEnumDef) {
      return value.label;
    } else
      return value.toString();
  }

  static String xddValueToString(Object? value, Iterable<XProp> colsOfXFdataItemToShow) {
    if (value is XFDataItem) {
      return xddValueToString_XFD(value, colsOfXFdataItemToShow);
    } else if (value is XEnumDef) {
      return value.label;
    } else
      return value.toString();
  }

  static String xddValueToString_XFD(XFDataItem value, Iterable<XProp> colsOfXFdataItemToShow) {
    var str = "";
    colsOfXFdataItemToShow.forEach((element) {
      str += (str.isEmpty ? "" : " ") + value[element.col_Name].toString();
    });
    return str;
  }
}

class XActionWidget extends StatefulWidget {
  XActionWidget(
    this.action, {
    this.onPressed,
    this.item,
    this.BTN_BorderColor = Colors.transparent,
    this.BTN_Color = const Color.fromARGB(255, 27, 27, 27),
    this.BTN_Height = 35,
    this.BTN_Width = 35,
    this.icon_Color,
    this.icon,
    this.child,
  }) : super();
  XAction action;
  dynamic item;
  Widget? child;
  Function()? onPressed;
  double? BTN_Width;
  double? BTN_Height;
  Color? BTN_Color;
  Color? BTN_BorderColor;
  Color? icon_Color;
  IconData? icon;

  @override
  _XActionWidgetState createState() => _XActionWidgetState();
}

class _XActionWidgetState extends State<XActionWidget> {
  @override
  void initState() {
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
        margin: EdgeInsets.only(top: 3, right: 5, bottom: 3),
        child: widget.action.child != null
            ? widget.action.child
            : XBtnbase(
                width: widget.BTN_Width,
                height: widget.BTN_Height,
                icon_WidthArea: 25,
                toolTip_Message: widget.action.toolTip,
                icon: widget.icon != null ? widget.icon : widget.action.icon ?? null,
                icon_Widget: widget.action.icon_Image != null
                    ? Image.memory(widget.action.icon_Image!)
                    : widget.action.icon == null && widget.icon == null
                        ? Text(widget.action.label.xSubString(0, 1).toUpperCase(), style: XStyles.xStyTextForLabel(textColor: Colors.white))
                        : null,
                icon_Size: 22,
                onPressed: () {
                  if (widget.action.cmd != null) {
                    setState(() {
                      widget.action.cmd!(widget.item, setState);
                    });
                  }
                  if (widget.onPressed != null) {
                    setState(() {
                      widget.onPressed!();
                    });
                  }
                },
                elevation: 0,
                decoration: BoxDecoration(color: widget.action.background_Color ?? widget.BTN_Color, border: Border.all(width: 1, color: widget.BTN_BorderColor ?? Colors.transparent), borderRadius: BorderRadius.all(Radius.circular(10))),
                label_Color: XColors.foregroundLight,
                icon_Color: widget.icon_Color != null ? widget.icon_Color : (widget.action.icon != null ? widget.action.icon_Color : null),
              ));
  }
}
